#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import types
from functools import wraps
from pyjava.api.mlsql import RayContext, PythonContext
from pyjava.storage import streaming_tar


def since(version):
    """
    A decorator that annotates a function to append the version of Spark the function was added.
    """
    import re
    indent_p = re.compile(r'\n( +)')

    def deco(f):
        indents = indent_p.findall(f.__doc__)
        indent = ' ' * (min(len(m) for m in indents) if indents else 0)
        f.__doc__ = f.__doc__.rstrip() + "\n\n%s.. versionadded:: %s" % (indent, version)
        return f

    return deco


def copy_func(f, name=None, sinceversion=None, doc=None):
    """
    Returns a function with same code, globals, defaults, closure, and
    name (or provide a new name).
    """
    # See
    # http://stackoverflow.com/questions/6527633/how-can-i-make-a-deepcopy-of-a-function-in-python
    fn = types.FunctionType(f.__code__, f.__globals__, name or f.__name__, f.__defaults__,
                            f.__closure__)
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    if doc is not None:
        fn.__doc__ = doc
    if sinceversion is not None:
        fn = since(sinceversion)(fn)
    return fn


def keyword_only(func):
    """
    A decorator that forces keyword arguments in the wrapped method
    and saves actual input keyword arguments in `_input_kwargs`.

    .. note:: Should only be used to wrap a method where first arg is `self`
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) > 0:
            raise TypeError("Method %s forces keyword arguments." % func.__name__)
        self._input_kwargs = kwargs
        return func(self, **kwargs)

    return wrapper


__all__ = ['RayContext', 'PythonContext', 'streaming_tar', 'context']



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function

from pyjava.api.mlsql import PythonContext
from pyjava.cache.code_cache import CodeCache
from pyjava.utils import *

# 'resource' is a Unix specific module.
has_resource_module = True
try:
    import resource
except ImportError:
    has_resource_module = False

import traceback
import logging

from pyjava.serializers import \
    write_with_length, \
    write_int, \
    read_int, read_bool, SpecialLengths, UTF8Deserializer, \
    PickleSerializer, ArrowStreamPandasSerializer, ArrowStreamSerializer

if sys.version >= '3':
    basestring = str
else:
    pass

# logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level='DEBUG')

pickleSer = PickleSerializer()
utf8_deserializer = UTF8Deserializer()

globals_namespace = globals()


def read_command(serializer, file):
    command = serializer.load_stream(file)
    return command


def chain(f, g):
    """chain two functions together """
    return lambda *a: g(f(*a))


def main(infile, outfile):
    try:
        try:
            import ray
        except ImportError:
            pass
        # set up memory limits
        memory_limit_mb = int(os.environ.get('PY_EXECUTOR_MEMORY', "-1"))
        if memory_limit_mb > 0 and has_resource_module:
            total_memory = resource.RLIMIT_AS
            try:
                (soft_limit, hard_limit) = resource.getrlimit(total_memory)
                msg = "Current mem limits: {0} of max {1}\n".format(soft_limit, hard_limit)
                logging.info(msg)

                # convert to bytes
                new_limit = memory_limit_mb * 1024 * 1024

                if soft_limit == resource.RLIM_INFINITY or new_limit < soft_limit:
                    msg = "Setting mem limits to {0} of max {1}\n".format(new_limit, new_limit)
                    logging.info(msg)
                    resource.setrlimit(total_memory, (new_limit, new_limit))

            except (resource.error, OSError, ValueError) as e:
                # not all systems support resource limits, so warn instead of failing
                logging.warning("WARN: Failed to set memory limit: {0}\n".format(e))
        split_index = read_int(infile)
        logging.info("split_index:%s" % split_index)
        if split_index == -1:  # for unit tests
            sys.exit(-1)

        is_barrier = read_bool(infile)
        bound_port = read_int(infile)
        logging.info(f"is_barrier {is_barrier}, port {bound_port}")

        conf = {}
        for i in range(read_int(infile)):
            k = utf8_deserializer.loads(infile)
            v = utf8_deserializer.loads(infile)
            conf[k] = v
            logging.debug(f"conf {k}:{v}")

        command = utf8_deserializer.loads(infile)
        ser = ArrowStreamSerializer()

        timezone = conf["timezone"] if "timezone" in conf else None

        out_ser = ArrowStreamPandasSerializer(timezone, True, True)
        is_interactive = os.environ.get('PY_INTERACTIVE', "no") == "yes"
        import uuid
        context_id = str(uuid.uuid4())

        if not os.path.exists(context_id):
            os.mkdir(context_id)

        def process():
            try:
                input_data = ser.load_stream(infile)
                code = CodeCache.get(command)
                if is_interactive:
                    global data_manager
                    global context
                    data_manager = PythonContext(context_id, input_data, conf)
                    context = data_manager
                    global globals_namespace
                    exec(code, globals_namespace, globals_namespace)
                else:
                    data_manager = PythonContext(context_id, input_data, conf)
                    n_local = {"data_manager": data_manager, "context": data_manager}
                    exec(code, n_local, n_local)
                out_iter = data_manager.output()
                write_int(SpecialLengths.START_ARROW_STREAM, outfile)
                out_ser.dump_stream(out_iter, outfile)
            finally:

                try:
                    import shutil
                    shutil.rmtree(context_id)
                except:
                    pass

                try:
                    if hasattr(out_iter, 'close'):
                        out_iter.close()
                except:
                    pass

                try:
                    del data_manager
                except:
                    pass

        process()


    except Exception:
        try:
            write_int(SpecialLengths.ARROW_STREAM_CRASH, outfile)
            write_int(SpecialLengths.PYTHON_EXCEPTION_THROWN, outfile)
            write_with_length(traceback.format_exc().encode("utf-8"), outfile)
        except IOError:
            # JVM close the socket
            pass
        except Exception:
            # Write the error to stderr if it happened while serializing
            print("Py worker failed with exception:", file=sys.stderr)
            print(traceback.format_exc(), file=sys.stderr)
        sys.exit(-1)

    write_int(SpecialLengths.END_OF_DATA_SECTION, outfile)
    flag = read_int(infile)
    if flag == SpecialLengths.END_OF_STREAM:
        write_int(SpecialLengths.END_OF_STREAM, outfile)
    else:
        # write a different value to tell JVM to not reuse this worker
        write_int(SpecialLengths.END_OF_DATA_SECTION, outfile)
        sys.exit(-1)


if __name__ == '__main__':
    # Read information about how to connect back to the JVM from the environment.
    java_port = int(os.environ["PYTHON_WORKER_FACTORY_PORT"])
    (sock_file, _) = local_connect_and_auth(java_port)
    main(sock_file, sock_file)



import logging
import os
import socket
import traceback

import ray

import pyjava.utils as utils
from pyjava.api.mlsql import RayContext
from pyjava.rayfix import RayWrapper
from pyjava.serializers import \
    write_with_length, \
    write_int, read_int, \
    SpecialLengths, ArrowStreamPandasSerializer


class SocketNotBindException(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message


class DataServerWithId(object):
    def __init__(self, host, port, server_id):
        self.host = host
        self.port = port
        self.server_id = server_id


class OnceServer(object):
    def __init__(self, host, port, timezone):
        self.host = host
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(5 * 60)
        self.out_ser = ArrowStreamPandasSerializer(timezone, False, None)
        self.is_bind = False
        self.is_dev = utils.is_dev()

    def bind(self):
        try:
            self.socket.bind((self.host, self.port))
            self.is_bind = True
            self.socket.listen(1)
        except Exception:
            print(traceback.format_exc())

        return self.socket.getsockname()

    def close(self):
        self.socket.close()

    def serve(self, data):
        from pyjava.api.mlsql import PythonContext
        if not self.is_bind:
            raise SocketNotBindException(
                "Please invoke server.bind() before invoke server.serve")
        conn, addr = self.socket.accept()
        sockfile = conn.makefile("rwb", int(
            os.environ.get("BUFFER_SIZE", 65536)))
        infile = sockfile  # os.fdopen(os.dup(conn.fileno()), "rb", 65536)
        out = sockfile  # os.fdopen(os.dup(conn.fileno()), "wb", 65536)
        try:
            write_int(SpecialLengths.START_ARROW_STREAM, out)
            out_data = ([df[name] for name in df] for df in
                        PythonContext.build_chunk_result(data, 1024))
            self.out_ser.dump_stream(out_data, out)

            write_int(SpecialLengths.END_OF_DATA_SECTION, out)
            write_int(SpecialLengths.END_OF_STREAM, out)
            out.flush()
            if self.is_dev:
                print("all data  in ray task have been consumed.")
            read_int(infile)
        except Exception:
            try:
                write_int(SpecialLengths.ARROW_STREAM_CRASH, out)
                ex = traceback.format_exc()
                print(ex)
                write_int(SpecialLengths.PYTHON_EXCEPTION_THROWN, out)
                write_with_length(ex.encode("utf-8"), out)
                out.flush()
                read_int(infile)
            except IOError:
                # JVM close the socket
                pass
            except Exception:
                # Write the error to stderr if it happened while serializing
                print("Py worker failed with exception:")
                print(traceback.format_exc())
                pass

        conn.close()


@ray.remote
class RayDataServer(object):

    def __init__(self, server_id, java_server, port=0, timezone="Asia/Harbin"):
        self.server = OnceServer(
            RayWrapper().get_address(), port, java_server.timezone)
        try:
            (rel_host, rel_port) = self.server.bind()
        except Exception as e:
            print(traceback.format_exc())
            raise e

        self.host = rel_host
        self.port = rel_port
        self.timezone = timezone
        self.server_id = server_id
        self.java_server = java_server
        self.is_dev = utils.is_dev()

    def serve(self, func_for_row=None, func_for_rows=None):
        try:
            if func_for_row is not None:
                data = (func_for_row(item)
                        for item in RayContext.fetch_once_as_rows(self.java_server))
            elif func_for_rows is not None:
                data = func_for_rows(
                    RayContext.fetch_once_as_rows(self.java_server))
            self.server.serve(data)
        except Exception as e:
            logging.error(f"Fail to processing data in  Ray Data Server {self.host}:{self.port}")
            raise e
        finally:
            self.close()

    def close(self):
        try:
            self.server.close()
        except Exception:
            print(traceback.format_exc())
        ray.actor.exit_actor()

    def connect_info(self):
        return DataServerWithId(self.host, self.port, self.server_id)



from typing import Any, NoReturn, Callable, Dict, List
import tempfile
import os

from pyjava.api.mlsql import PythonContext


class Utils(object):

    @staticmethod
    def show_plt(plt: Any, context: PythonContext) -> NoReturn:
        content = Utils.gen_img(plt)
        context.build_result([{"content": content, "mime": "image"}])

    @staticmethod
    def gen_img(plt: Any) -> str:
        import base64
        import uuid
        img_path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()) + ".png")
        plt.savefig(img_path)
        with open(img_path, mode='rb') as file:
            file_content = base64.b64encode(file.read()).decode()
        return file_content



import io
import os
import tarfile
import tempfile

import uuid

BLOCK_SIZE = 1024 * 64


class FileStream(object):
    def __init__(self):
        self.buffer = io.BytesIO()
        self.offset = 0

    def write(self, s):
        self.buffer.write(s)
        self.offset += len(s)

    def tell(self):
        return self.offset

    def close(self):
        self.buffer.close()

    def pop(self):
        s = self.buffer.getvalue()
        self.buffer.close()
        self.buffer = io.BytesIO()
        return s


def build_rows_from_file(target_dir):
    file = FileStream()
    start = 0
    for i in stream_build_tar(target_dir, file):
        v = file.pop()
        offset = len(v)
        if len(v) > 0:
            yield {"start": start, "offset": offset, "value": v}
            start = start + offset


def save_rows_as_file(data, target_dir):
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    tf_path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
    with open(tf_path, "wb") as tf:
        for block_row in data:
            tf.write(block_row["value"])
    with open(tf_path, "rb") as tf:
        tt = tarfile.open(tf.name, mode="r:")
        tt.extractall(target_dir)
        tt.close()
    os.remove(tf_path)


def stream_build_tar(file_dir, streaming_fp):
    tar = tarfile.TarFile.open(fileobj=streaming_fp, mode="w:")
    for root, dirs, files in os.walk(file_dir):
        for in_filename in files:
            stat = os.stat(os.path.join(root, in_filename))
            tar_info = tarfile.TarInfo(os.path.join(root.lstrip(file_dir), in_filename))
            # tar_info.path = root.lstrip(file_dir)
            tar_info.mtime = stat.st_mtime
            tar_info.size = stat.st_size

            tar.addfile(tar_info)

            yield

            with open(os.path.join(root, in_filename), 'rb') as in_fp:
                # total_size = 0
                while True:
                    s = in_fp.read(BLOCK_SIZE)

                    if len(s) > 0:
                        tar.fileobj.write(s)
                        yield

                    if len(s) < BLOCK_SIZE:
                        blocks, remainder = divmod(tar_info.size, tarfile.BLOCKSIZE)

                        if remainder > 0:
                            tar.fileobj.write(tarfile.NUL *
                                              (tarfile.BLOCKSIZE - remainder))

                            yield

                            blocks += 1

                        tar.offset += blocks * tarfile.BLOCKSIZE
                        break
    tar.close()


def main():
    rows = build_rows_from_file("/Users/allwefantasy/data/mlsql/homes/demo/tmp/minist_model")
    save_rows_as_file(rows, "/Users/allwefantasy/data/mlsql/homes/demo/tmp/minist_model3")


if __name__ == '__main__':
    main()



from typing import Any, Generic, List, Callable, Union, Tuple, Iterable

import ray
from pyjava.api.mlsql import RayContext


class KoloRawDatasource(object):
    def __init__(self, data_refs: List[str], num_tables_per_block: int = 1):
        self.data_refs = data_refs
        self.block_refs = []

        @ray.remote
        def make_block(_data_ref):
            block_refs = []
            data_iter = RayContext.fetch_data_from_single_data_server_as_arrow(_data_ref)
            temp_box = []
            for arrow_table in data_iter:
                temp_box.append(arrow_table)
                if len(temp_box) == num_tables_per_block:
                    for t in temp_box:
                        block_refs.append(ray.put(t))
                    temp_box.clear()
            if len(temp_box) != 0:
                for t in temp_box:
                    block_refs.append(ray.put(t))
                temp_box.clear()
            return block_refs

        for data_ref in data_refs:
            for item in ray.get(make_block.remote(data_ref)):
                self.block_refs.append(item)

    def to_dataset(self):
        return ray.data.from_arrow_refs(self.block_refs)



import logging
import os
import socket
from distutils.version import StrictVersion
import uuid
from urllib.parse import quote
import json

import pandas as pd
import sys
import pyjava.utils as utils
import requests

from pyjava.serializers import ArrowStreamSerializer
from pyjava.serializers import read_int
from pyjava.utils import utf8_deserializer
from pyjava.storage import streaming_tar

if sys.version >= '3':
    basestring = str
else:
    pass


class DataServer(object):
    def __init__(self, host, port, timezone):
        self.host = host
        self.port = port
        self.timezone = timezone


class LogClient(object):
    def __init__(self, conf):
        self.conf = conf
        if 'spark.mlsql.log.driver.url' in self.conf:
            self.url = self.conf['spark.mlsql.log.driver.url']
            self.log_user = self.conf['PY_EXECUTE_USER']
            self.log_token = self.conf['spark.mlsql.log.driver.token']
            self.log_group_id = self.conf['groupId']

    def log_to_driver(self, msg):
        if 'spark.mlsql.log.driver.url' not in self.conf:
            if self.conf['PY_EXECUTE_USER'] and self.conf['groupId']:
                logging.info("[owner] [{}] [groupId] [{}] __MMMMMM__ {}".format(self.conf['PY_EXECUTE_USER'],
                                                                                self.conf['groupId'], msg))
            else:
                logging.info(msg)
            return
        resp = json.dumps(
            {"sendLog": {
                "token": self.log_token,
                "logLine": "[owner] [{}] [groupId] [{}] __MMMMMM__ {}".format(self.log_user, self.log_group_id, msg)
            }}, ensure_ascii=False)
        requests.post(self.url, data=resp, headers={'Content-Type': 'application/json;charset=UTF-8'})

    def close(self):
        if hasattr(self, "conn"):
            self.conn.close()
            self.conn = None


class PythonContext(object):
    cache = {}

    def __init__(self, context_id, iterator, conf):
        self.context_id = context_id
        self.data_mmap_file_ref = {}
        self.input_data = iterator
        self.output_data = [[]]
        self.conf = conf
        self.schema = ""
        self.have_fetched = False
        self.log_client = LogClient(self.conf)
        if "pythonMode" in conf and conf["pythonMode"] == "ray":
            self.rayContext = RayContext(self)

    def set_output(self, value, schema=""):
        self.output_data = value
        self.schema = schema

    @staticmethod
    def build_chunk_result(items, block_size=1024):
        buffer = []
        for item in items:
            buffer.append(item)
            if len(buffer) == block_size:
                df = pd.DataFrame(buffer, columns=buffer[0].keys())
                buffer.clear()
                yield df

        if len(buffer) > 0:
            df = pd.DataFrame(buffer, columns=buffer[0].keys())
            buffer.clear()
            yield df

    def build_result(self, items, block_size=1024):
        self.output_data = ([df[name] for name in df]
                            for df in PythonContext.build_chunk_result(items, block_size))

    def build_result_from_dir(self, target_dir, block_size=1024):
        items = streaming_tar.build_rows_from_file(target_dir)
        self.build_result(items, block_size)

    def output(self):
        return self.output_data

    def __del__(self):
        logging.info("==clean== context")
        if self.log_client is not None:
            try:
                self.log_client.close()
            except Exception as e:
                pass

        if 'data_mmap_file_ref' in self.data_mmap_file_ref:
            try:
                self.data_mmap_file_ref['data_mmap_file_ref'].close()
            except Exception as e:
                pass

    def noops_fetch(self):
        for item in self.fetch_once():
            pass

    def fetch_once_as_dataframe(self):
        for df in self.fetch_once():
            yield df

    def fetch_once_as_rows(self):
        for df in self.fetch_once_as_dataframe():
            for row in df.to_dict('records'):
                yield row

    def fetch_once_as_batch_rows(self):
        for df in self.fetch_once_as_dataframe():
            yield (row for row in df.to_dict('records'))

    def fetch_once(self):
        import pyarrow as pa
        if self.have_fetched:
            raise Exception("input data can only be fetched once")
        self.have_fetched = True
        for items in self.input_data:
            yield pa.Table.from_batches([items]).to_pandas()

    def fetch_as_dir(self, target_dir):
        if len(self.data_servers()) > 1:
            raise Exception("Please make sure you have only one partition on Java/Spark Side")
        items = self.fetch_once_as_rows()
        streaming_tar.save_rows_as_file(items, target_dir)


class PythonProjectContext(object):
    def __init__(self):
        self.params_read = False
        self.conf = {}
        self.read_params_once()
        self.log_client = LogClient(self.conf)

    def read_params_once(self):
        if not self.params_read:
            self.params_read = True
            infile = sys.stdin.buffer
            for i in range(read_int(infile)):
                k = utf8_deserializer.loads(infile)
                v = utf8_deserializer.loads(infile)
                self.conf[k] = v

    def input_data_dir(self):
        return self.conf["tempDataLocalPath"]

    def output_model_dir(self):
        return self.conf["tempModelLocalPath"]

    def __del__(self):
        self.log_client.close()


class RayContext(object):
    cache = {}
    conn_cache = {}

    def __init__(self, python_context):
        self.python_context = python_context
        self.servers = []
        self.server_ids_in_ray = []
        self.is_setup = False
        self.is_dev = utils.is_dev()
        self.is_in_mlsql = True
        self.mock_data = []
        if "directData" not in python_context.conf:
            for item in self.python_context.fetch_once_as_rows():
                self.server_ids_in_ray.append(str(uuid.uuid4()))
                self.servers.append(DataServer(
                    item["host"], int(item["port"]), item["timezone"]))

    def data_servers(self):
        return self.servers

    def conf(self):
        return self.python_context.conf

    def data_servers_in_ray(self):
        import ray
        from pyjava.rayfix import RayWrapper
        rayw = RayWrapper()
        for server_id in self.server_ids_in_ray:
            server = rayw.get_actor(server_id)
            yield ray.get(server.connect_info.remote())

    def build_servers_in_ray(self):
        from pyjava.rayfix import RayWrapper
        from pyjava.api.serve import RayDataServer
        import ray
        buffer = []
        rayw = RayWrapper()
        for (server_id, java_server) in zip(self.server_ids_in_ray, self.servers):
            # rds = RayDataServer.options(name=server_id, detached=True, max_concurrency=2).remote(server_id, java_server,
            #                                                                                0,
            #                                                                                java_server.timezone)
            rds = rayw.options(RayDataServer, name=server_id, detached=True, max_concurrency=2).remote(server_id,
                                                                                                       java_server,
                                                                                                       0,
                                                                                                       java_server.timezone)
            res = ray.get(rds.connect_info.remote())
            logging.debug("build ray data server server_id:{} java_server: {} servers:{}".format(server_id,
                                                                                                 str(vars(
                                                                                                     java_server)),
                                                                                                 str(vars(res))))
            buffer.append(res)
        return buffer

    @staticmethod
    def connect(_context, url, **kwargs):
        if isinstance(_context, PythonContext):
            context = _context
        elif isinstance(_context, dict):
            if 'context' in _context:
                context = _context['context']
            else:
                '''
                we are not in MLSQL
                '''
                context = PythonContext("", [], {"pythonMode": "ray"})
                context.rayContext.is_in_mlsql = False
        else:
            raise Exception("context is not detect. make sure it's in globals().")

        if url == "local":
            from pyjava.rayfix import RayWrapper
            ray = RayWrapper()
            if ray.ray_version < StrictVersion('1.6.0'):
                raise Exception("URL:local is only support in ray >= 1.6.0")
            # if not ray.ray_instance.is_initialized:
            ray.ray_instance.shutdown()
            ray.ray_instance.init(namespace="default")

        elif url is not None:
            from pyjava.rayfix import RayWrapper
            ray = RayWrapper()
            is_udf_client = context.conf.get("UDF_CLIENT")
            if is_udf_client is None:
                ray.shutdown()
                ray.init(url, **kwargs)
            if is_udf_client and url not in RayContext.conn_cache:
                ray.init(url, **kwargs)
                RayContext.conn_cache[url] = 1

        return context.rayContext

    def setup(self, func_for_row, func_for_rows=None):
        if self.is_setup:
            raise ValueError("setup can be only invoke once")
        self.is_setup = True

        is_data_mode = "dataMode" in self.conf() and self.conf()["dataMode"] == "data"

        if not is_data_mode:
            raise Exception('''
Please setup dataMode as data instead of model. 
Try run: `!python conf "dataMode=data"` or 
add comment like: `#%dataMode=data` if you are in notebook.
            ''')

        import ray
        from pyjava.rayfix import RayWrapper
        rayw = RayWrapper()

        if not self.is_in_mlsql:
            if func_for_rows is not None:
                func = ray.remote(func_for_rows)
                return ray.get(func.remote(self.mock_data))
            else:
                func = ray.remote(func_for_row)

                def iter_all(rows):
                    return [ray.get(func.remote(row)) for row in rows]

                iter_all_func = ray.remote(iter_all)
                return ray.get(iter_all_func.remote(self.mock_data))

        buffer = []
        for server_info in self.build_servers_in_ray():
            server = rayw.get_actor(server_info.server_id)
            rci = ray.get(server.connect_info.remote())
            buffer.append(rci)
            server.serve.remote(func_for_row, func_for_rows)
        items = [vars(server) for server in buffer]
        self.python_context.build_result(items, 1024)
        return buffer

    def foreach(self, func_for_row):
        return self.setup(func_for_row)

    def map_iter(self, func_for_rows):
        return self.setup(None, func_for_rows)

    def collect(self):
        for shard in self.data_servers():
            for row in RayContext.fetch_once_as_rows(shard):
                yield row

    def fetch_as_dir(self, target_dir, servers=None):
        if not servers:
            servers = self.data_servers()
        if len(servers) > 1:
            raise Exception("Please make sure you have only one partition on Java/Spark Side")

        items = RayContext.collect_from(servers)
        streaming_tar.save_rows_as_file(items, target_dir)

    def build_result(self, items, block_size=1024):
        self.python_context.build_result(items, block_size)

    def build_result_from_dataframe(self, df):
        rows = (row for row in df.to_dict('records'))
        self.python_context.build_result(rows)

    def build_result_from_dir(self, target_path):
        self.python_context.build_result_from_dir(target_path)

    @staticmethod
    def parse_servers(host_ports):
        hosts = host_ports.split(",")
        hosts = [item.split(":") for item in hosts]
        return [DataServer(item[0], int(item[1]), "") for item in hosts]

    @staticmethod
    def fetch_as_repeatable_file(context_id, data_servers, file_ref, batch_size):
        import pyarrow as pa

        def inner_fetch():
            for data_server in data_servers:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                    out_ser = ArrowStreamSerializer()
                    sock.connect((data_server.host, data_server.port))
                    buffer_size = int(os.environ.get("BUFFER_SIZE", 65536))
                    infile = os.fdopen(os.dup(sock.fileno()), "rb", buffer_size)
                    result = out_ser.load_stream(infile)
                    for batch in result:
                        yield batch

        def gen_by_batch():
            import numpy as np
            import math
            if 'data_mmap_file_ref' not in file_ref:
                file_ref['data_mmap_file_ref'] = pa.memory_map(context_id + "/__input__.dat")
            reader = pa.ipc.open_file(file_ref['data_mmap_file_ref'])
            num_record_batches = reader.num_record_batches
            for i in range(num_record_batches):
                df = reader.get_batch(i).to_pandas()
                for small_batch in np.array_split(df, math.floor(df.shape[0] / batch_size)):
                    yield small_batch

        if 'data_mmap_file_ref' in file_ref:
            return gen_by_batch()
        else:
            writer = None
            for batch in inner_fetch():
                if writer is None:
                    writer = pa.RecordBatchFileWriter(context_id + "/__input__.dat", batch.schema)
                writer.write_batch(batch)
            writer.close()
            return gen_by_batch()

    def collect_as_file(self, batch_size):
        data_servers = self.data_servers()
        python_context = self.python_context
        return RayContext.fetch_as_repeatable_file(python_context.context_id, data_servers,
                                                   python_context.data_mmap_file_ref,
                                                   batch_size)

    @staticmethod
    def collect_from(servers):
        for shard in servers:
            for row in RayContext.fetch_once_as_rows(shard):
                yield row

    def to_pandas(self):
        items = [row for row in self.collect()]
        return pd.DataFrame(data=items)

    def to_dataset(self, num_tables_per_block: int = 1):
        from pyjava.data.datasource import KoloRawDatasource
        source = KoloRawDatasource(self.servers, num_tables_per_block)
        return source.to_dataset()

    @staticmethod
    def fetch_once_as_rows(data_server):
        for df in RayContext.fetch_data_from_single_data_server(data_server):
            for row in df.to_dict('records'):
                yield row

    @staticmethod
    def fetch_data_from_single_data_server(data_server):
        for table in RayContext.fetch_data_from_single_data_server_as_arrow(data_server):
            yield table.to_pandas()

    @staticmethod
    def fetch_data_from_single_data_server_as_arrow(data_server):
        out_ser = ArrowStreamSerializer()
        import pyarrow as pa
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.connect((data_server.host, data_server.port))
            buffer_size = int(os.environ.get("BUFFER_SIZE", 65536))
            infile = sock.makefile("rb", buffer_size)  # os.fdopen(os.dup(sock.fileno()), "rb", buffer_size)
            result = out_ser.load_stream(infile)
            for items in result:
                yield pa.Table.from_batches([items])



from distutils.version import StrictVersion

import ray
import logging


def last(func):
    func.__module__ = "pyjava_auto_generate__exec__"
    return func


class RayWrapper:

    def __init__(self):
        self.ray_version = "2.0.0" if "dev" in ray.__version__ else StrictVersion(ray.__version__)
        self.ray_instance = ray

    def __getattr__(self, attr):
        return getattr(self.ray_instance, attr)

    def get_address(self):
        if self.ray_version >= StrictVersion('1.0.0'):
            return ray.get_runtime_context().worker.node_ip_address
        else:
            return ray.services.get_node_ip_address()

    def init(self, address, **kwargs):
        logging.debug(f"address {address} {kwargs}")
        if self.ray_version >= StrictVersion('1.4.0'):
            if "namespace" in kwargs.keys():
                ray.util.connect(conn_str=address, **kwargs)
            else:
                ray.util.connect(conn_str=address, namespace="default", **kwargs)
        elif self.ray_version >= StrictVersion('1.0.0'):
            logging.debug(f"try to connect to ray {address}")
            ray.util.connect(conn_str=address, **kwargs)
        elif self.ray_version == StrictVersion('0.8.7'):
            ray.init(address=address, **kwargs)
        else:
            ray.init(redis_address=address, **kwargs)

    def shutdown(self):
        if self.ray_version >= StrictVersion('1.0.0'):
            try:
                ray.util.disconnect()
            except Exception as e:
                pass
        else:
            ray.shutdown(exiting_interpreter=False)

    def options(self, actor_class, **kwargs):
        if 'detached' in kwargs and self.ray_version >= StrictVersion('1.0.0'):
            del kwargs['detached']
            kwargs['lifetime'] = 'detached'
        logging.debug(f"actor build options: {kwargs}")
        return actor_class.options(**kwargs)

    def get_actor(self, name):
        if self.ray_version >= StrictVersion('1.0.0'):
            return ray.get_actor(name)
        elif self.ray_version == StrictVersion('0.8.7'):
            return ray.get_actor(name)
        else:
            return ray.experimental.get_actor(name)



import uuid
from typing import Any, NoReturn, Callable, Dict, List
import ray
import time
from ray.util.client.common import ClientActorHandle, ClientObjectRef

from pyjava.api.mlsql import RayContext
from pyjava.storage import streaming_tar


@ray.remote
class UDFMaster(object):
    def __init__(self, num: int, conf: Dict[str, str],
                 init_func: Callable[[List[ClientObjectRef], Dict[str, str]], Any],
                 apply_func: Callable[[Any, Any], Any]):
        model_servers = RayContext.parse_servers(conf["modelServers"])
        items = RayContext.collect_from(model_servers)
        model_refs = [ray.put(item) for item in items]

        udfWorkerConf = {}

        if "num_cpus" in conf:
            udfWorkerConf["num_cpus"] = int(conf["num_cpus"])

        if "num_gpus" in conf:
            udfWorkerConf["num_gpus"] = int(conf["num_gpus"])

        custom_resources = [(key.split("resource.")[1], float(conf[key])) for key in
                            conf.keys() if
                            key.startswith("resource.")]

        if len(custom_resources) > 0:
            udfWorkerConf["resources"] = dict(custom_resources)

        self.actors = dict(
            [(index, UDFWorker.options(**udfWorkerConf).remote(model_refs, conf, init_func, apply_func)) for index in range(num)])
        self._idle_actors = [index for index in range(num)]

    def get(self) -> List[Any]:
        while len(self._idle_actors) == 0:
            time.sleep(0.001)
        index = self._idle_actors.pop()
        return [index, self.actors[index]]

    def give_back(self, v) -> NoReturn:
        self._idle_actors.append(v)

    def shutdown(self) -> NoReturn:
        [ray.kill(self.actors[index]) for index in self._idle_actors]


@ray.remote
class UDFWorker(object):
    def __init__(self,
                 model_refs: List[ClientObjectRef],
                 conf: Dict[str, str],
                 init_func: Callable[[List[ClientObjectRef], Dict[str, str]], Any],
                 apply_func: Callable[[Any, Any], Any]):
        self.model = init_func(model_refs, conf)
        self.apply_func = apply_func

    def apply(self, v: Any) -> Any:
        return self.apply_func(self.model, v)

    def shutdown(self):
        ray.actor.exit_actor()


class UDFBuilder(object):
    @staticmethod
    def build(ray_context: RayContext,
              init_func: Callable[[List[ClientObjectRef], Dict[str, str]], Any],
              apply_func: Callable[[Any, Any], Any]) -> NoReturn:
        conf = ray_context.conf()
        udf_name = conf["UDF_CLIENT"]
        max_concurrency = int(conf.get("maxConcurrency", "3"))

        try:
            stop_flag = True
            counter = 30
            temp_udf_master = ray.get_actor(udf_name)
            ray.kill(temp_udf_master)
            while stop_flag or counter < 0:
                time.sleep(1)
                try:
                    ray.get_actor(udf_name)
                except Exception:
                    stop_flag = False
                    counter = counter - 1

        except Exception as inst:
            print(inst)
            pass

        UDFMaster.options(name=udf_name, lifetime="detached", max_concurrency=max_concurrency).remote(
            max_concurrency, conf, init_func, apply_func)
        ray_context.build_result([])

    @staticmethod
    def apply(ray_context: RayContext):
        conf = ray_context.conf()
        udf_name = conf["UDF_CLIENT"]
        udf_master = ray.get_actor(udf_name)
        [index, worker] = ray.get(udf_master.get.remote())
        input_value = [row["value"] for row in ray_context.python_context.fetch_once_as_rows()]
        try:
            res = ray.get(worker.apply.remote(input_value))
        except Exception as inst:
            res = []
            print(inst)
        udf_master.give_back.remote(index)
        ray_context.build_result([res])


class UDFBuildInFunc(object):
    @staticmethod
    def init_tf(model_refs: List[ClientObjectRef], conf: Dict[str, str]) -> Any:
        from tensorflow.keras import models
        model_path = "./tmp/model/{}".format(str(uuid.uuid4()))
        streaming_tar.save_rows_as_file((ray.get(ref) for ref in model_refs), model_path)
        return models.load_model(model_path)



#!/usr/bin/env python

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys

from setuptools import setup

if sys.version_info < (3, 6):
    print("Python versions prior to 2.7 are not supported for pip installed pyjava.",
          file=sys.stderr)
    sys.exit(-1)

try:
    exec(open('pyjava/version.py').read())
except IOError:
    print("Failed to load PyJava version file for packaging. You must be in PyJava's python dir.",
          file=sys.stderr)
    sys.exit(-1)
VERSION = __version__  # noqa
# A temporary path so we can access above the Python project root and fetch scripts and jars we need
TEMP_PATH = "deps"

# Provide guidance about how to use setup.py
incorrect_invocation_message = """
If you are installing pyjava from source, you must first build 
run sdist.
    Building the source dist is done in the Python directory:
      cd python
      python setup.py sdist
      pip install dist/*.tar.gz"""

try:
    setup(
        name='pyjava',
        version=VERSION,
        description='PyJava Python API',
        long_description="",
        author='allwefantasy',
        author_email='allwefantasy@gmail.com',
        url='https://github.com/allwefantasy/pyjava',
        packages=['pyjava',
                  'pyjava.api',
                  'pyjava.udf',
                  'pyjava.datatype',
                  'pyjava.storage',
                  'pyjava.cache',
                  'pyjava.data'
                  ],
        include_package_data=True,
        package_dir={
            'pyjava.sbin': 'deps/sbin'
        },
        package_data={
            'pyjava.sbin': []},
        license='http://www.apache.org/licenses/LICENSE-2.0',
        setup_requires=['pypandoc'],
        extras_require={
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy']
    )
finally:
    print("--------")



import os
import sys

from pyjava.serializers import write_with_length, UTF8Deserializer, \
    PickleSerializer

if sys.version >= '3':
    basestring = str
else:
    pass

pickleSer = PickleSerializer()
utf8_deserializer = UTF8Deserializer()


def is_dev():
    return 'MLSQL_DEV' in os.environ and int(os.environ["MLSQL_DEV"]) == 1


def require_minimum_pandas_version():
    """ Raise ImportError if minimum version of Pandas is not installed
    """
    # TODO(HyukjinKwon): Relocate and deduplicate the version specification.
    minimum_pandas_version = "0.23.2"

    from distutils.version import LooseVersion
    try:
        import pandas
        have_pandas = True
    except ImportError:
        have_pandas = False
    if not have_pandas:
        raise ImportError("Pandas >= %s must be installed; however, "
                          "it was not found." % minimum_pandas_version)
    if LooseVersion(pandas.__version__) < LooseVersion(minimum_pandas_version):
        raise ImportError("Pandas >= %s must be installed; however, "
                          "your version was %s." % (minimum_pandas_version, pandas.__version__))


def require_minimum_pyarrow_version():
    """ Raise ImportError if minimum version of pyarrow is not installed
    """
    # TODO(HyukjinKwon): Relocate and deduplicate the version specification.
    minimum_pyarrow_version = "0.12.1"

    from distutils.version import LooseVersion
    try:
        import pyarrow
        have_arrow = True
    except ImportError:
        have_arrow = False
    if not have_arrow:
        raise ImportError("PyArrow >= %s must be installed; however, "
                          "it was not found." % minimum_pyarrow_version)
    if LooseVersion(pyarrow.__version__) < LooseVersion(minimum_pyarrow_version):
        raise ImportError("PyArrow >= %s must be installed; however, "
                          "your version was %s." % (minimum_pyarrow_version, pyarrow.__version__))


def _exception_message(excp):
    """Return the message from an exception as either a str or unicode object.  Supports both
    Python 2 and Python 3.

    >>> msg = "Exception message"
    >>> excp = Exception(msg)
    >>> msg == _exception_message(excp)
    True

    >>> msg = u"unicöde"
    >>> excp = Exception(msg)
    >>> msg == _exception_message(excp)
    True
    """
    if hasattr(excp, "message"):
        return excp.message
    return str(excp)


def _do_server_auth(conn, auth_secret):
    """
    Performs the authentication protocol defined by the SocketAuthHelper class on the given
    file-like object 'conn'.
    """
    write_with_length(auth_secret.encode("utf-8"), conn)
    conn.flush()
    reply = UTF8Deserializer().loads(conn)
    if reply != "ok":
        conn.close()
        raise Exception("Unexpected reply from iterator server.")


def local_connect_and_auth(port):
    """
    Connect to local host, authenticate with it, and return a (sockfile,sock) for that connection.
    Handles IPV4 & IPV6, does some error handling.
    :param port
    :param auth_secret
    :return: a tuple with (sockfile, sock)
    """
    import socket
    sock = None
    errors = []
    # Support for both IPv4 and IPv6.
    # On most of IPv6-ready systems, IPv6 will take precedence.
    for res in socket.getaddrinfo("127.0.0.1", port, socket.AF_UNSPEC, socket.SOCK_STREAM):
        af, socktype, proto, _, sa = res
        try:
            sock = socket.socket(af, socktype, proto)
            sock.settimeout(15)
            sock.connect(sa)
            sockfile = sock.makefile("rwb", int(os.environ.get("BUFFER_SIZE", 65536)))
            return (sockfile, sock)
        except socket.error as e:
            emsg = _exception_message(e)
            errors.append("tried to connect to %s, but an error occured: %s" % (sa, emsg))
            sock.close()
            sock = None
    raise Exception("could not open socket: %s" % errors)



class CodeCache(object):
    cache = {}
    cache_max_size = 10 * 10000

    @staticmethod
    def get(code):
        if code in CodeCache.cache:
            return CodeCache.cache[code]
        else:
            return CodeCache.gen_cache(code)

    @staticmethod
    def gen_cache(code):
        if len(CodeCache.cache) > CodeCache.cache_max_size:
            CodeCache.cache.clear()
        CodeCache.cache[code] = compile(code, '<string>', 'exec')
        return CodeCache.cache[code]
